/*
 * Generation of server dispatch and handle functions.
 *
 * Copyright (C) 2019, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "common.h"
#include "dispatch.h"
#include "templates.h"



/* Generate each dispatch possibility within an interface. */

static void write_dispatcher_cases(FILE *fp, struct interface *iface,
                                   int compound)
{
  struct signature *sig;
  char *opcode, *opname, *protocol, *prefix, *ref, *s;

  for (sig = iface->signatures; sig != NULL; sig = sig->tail)
  {
    opname = get_operation_name(iface, sig);
    opcode = get_opcode_identifier(NULL, opname);
    prefix = get_operation_wrapper_prefix(sig->attributes);

    /* Convert to any base interface. */

    ref = get_object_conversion(iface, compound);

    /* Generate a reply if appropriate. */

    if (have_attribute(sig->attributes, "completion"))
      s = dispatch_function_wrapper_case;
    else
      s = dispatch_function_reply_wrapper_case;

    /* Generate the case and invocation. */

    fprintf(fp, s, opcode, prefix, opname, ref);

    /* Free allocated strings. */

    free(opcode);
    free(opname);
    free(ref);
  }
}

/* Dispatch to each interface-level function where the same protocol applies to
   an entire interface. Dispatch to each operation where an interface does not
   specify a protocol. */

static void write_dispatcher_interface_cases(FILE *fp, struct interface *iface,
                                             int by_protocol)
{
  struct interface_ref *base;
  char *protocol, *ref;

  for (base = iface->bases; base != NULL; base = base->tail)
  {
    protocol = get_protocol(base->iface->attributes);

    /* Write cases for each interface, using the protocol to select between the
       cases. */

    if (by_protocol && (protocol != NULL))
    {
      /* Convert to any base interface. */

      ref = get_object_conversion(base->iface, 1);
      fprintf(fp, dispatch_function_interface_case, protocol, base->name, ref);
      free(ref);
    }

    /* Write cases for each operation provided by the interface so that they can
       be selected. Note that this requires distinct opcodes to be used, so in
       practice requiring explicit opcodes to be indicated. */

    else if (!by_protocol && (protocol == NULL))
      write_dispatcher_cases(fp, base->iface, 1);
  }
}

/* Return whether any base interface specifies a protocol. */

static int have_interfaces_using_protocols(struct interface *iface)
{
  struct interface_ref *base;

  for (base = iface->bases; base != NULL; base = base->tail)
    if (get_protocol(base->iface->attributes) != NULL)
      return 1;

  return 0;
}



/* Generate dispatch function signatures. */

void write_dispatcher_signature(const char *name, enum signature_role role,
                                FILE *fp)
{
  fprintf(fp, dispatch_function_signature, name, name);
  fputs(get_signature_terminator(role), fp);
}

/* Generate handler function signatures. */

void write_handler_signature(const char *name, enum signature_role role,
                             FILE *fp)
{
  fprintf(fp, handle_function_signature, name, name);
  fputs(get_signature_terminator(role), fp);
}

/* Generate a dispatch function for the different operations. */

void write_dispatcher(FILE *fp, struct interface *iface)
{
  char *protocol = get_protocol(iface->attributes);

  write_dispatcher_signature(iface->name, DEFINITION_ROLE, fp);

  /* Declare an error variable to support testing for already-sent messages. */

  fputs("  long err;\n\n", fp);

  /* Without a protocol applying to the entire interface, dispatch using the
     protocol from the message label. */

  if (protocol == NULL)
  {
    fputs("  switch (l4_msgtag_label(msg->tag))\n  {\n", fp);

    /* Dispatch using the protocol to base interfaces employing protocols. */

    if (have_interfaces_using_protocols(iface))
      write_dispatcher_interface_cases(fp, iface, 1);
  }

  /* If a protocol applies to the entire interface, dispatch using an operation
     indicator in the word data. */

  else
    fputs("  switch (ipc_message_get_word(msg, 0))\n  {\n", fp);

  /* Dispatch to operations provided by compound interfaces. */

  write_dispatcher_interface_cases(fp, iface, 0);

  /* Dispatch to each operation defined at this level. */

  write_dispatcher_cases(fp, iface, is_compound_interface(iface));

  /* Terminate the dispatcher. */

  fputs(dispatch_function_default_case, fp);
  fputs(dispatch_function_dispatcher_epilogue, fp);

  fputs(END_FUNCTION, fp);
}
